(* link.ml : batit le fichier main.k2 dans /tmp *)
(* version 0.63 *)
(* Rgis Cridlig 1994 *)

#open "sys";;
#open "misc";;
#open "filename";;
#open "config";;
#open "hashtable";;

let std_modules = ["io"; "eq"; "int"; "float"; "pair"; "list";
                   "fvect"; "fchar"; "fstring"; "exc"; "stream";
                   "arg"; "filename"; "hashtbl"; "lexing"; "parsing"; 
                   "printexc"; "printf"; "queue"; "random"; "sort"; 
                   "stack"; "sys"; "baltree"; "genlex"; "map"; "set"; "unix"]
;;

let default_used_implementations =
  ["none",     [];
   "fast",     std_modules;
   "cautious", std_modules @ ["vect"; "char"; "string"]]
;;

let exc_table = new_hashtable 31;;

let find_deps obj =
  let file = find_in_path (obj ^ ".ko")
  in let ic = open_in_bin file 
  in try
       let (deps:string list) = input_value ic
       in let (t:(int*string) hashtable) = input_value ic
       in do_hashtable (function e -> add_unique_to_assoctable exc_table e)
      	       	       t;
       deps
     with End_of_file | Failure _ ->
       close_in ic;
       prerr_begline " Corrupted dependence file ";
       prerr_endline file;
       raise Toplevel
;;

let rec toposort = fun
  [] res -> rev res
| modeps res -> 
    let rec find_min = function
                []     -> failwith "link: toposort"
     | (modu,[])::_    -> modu
     |         _::rest -> find_min rest
    and remove = fun
       _ [] res -> res      
     | modu ((m,deps)::l) res -> 
	if m=modu then remove modu l res
	else remove modu l ((m,except modu deps)::res)
    in let modu = find_min modeps
       in toposort (remove modu modeps []) (modu::res)
;;

let rec checklib = fun (* verification des modules de la lib standard *)
   _       []      modeps -> toposort modeps []
| libs (modu::mods) modeps ->
    begin
      let deps = find_deps modu
      and modsr = ref mods in
      do_list (function dep -> 
	        if not(mem dep mods) then
		  if not(mem dep (map fst modeps)) then
		    if mem dep libs then modsr := dep::!modsr
				    else fatal_error modu)
	     deps;
      checklib libs !modsr ((modu,deps)::modeps)
    end
;;

let rec check = fun (* verification des objets a linker *)
    libs     []      mods -> checklib libs (union ["io";"sys"] mods) []
  | libs (obj::objs) mods ->
      let deps = find_deps obj 
      and modsr = ref mods in
      do_list (function dep -> 
	        if not(mem dep (map (fun f -> basename f) objs)) then
		  if mem dep libs then modsr := union !modsr [dep]
				  else fatal_error (basename obj))
	     deps;
      check libs objs !modsr
;;

let link lib objects exec =
    let z2k2_main = path_tmp ^ "#camlot_" ^ (basename exec) in
    let main_k2 = open_out (z2k2_main ^ ".k2")
    and modules = check (assoc lib default_used_implementations) 
			(rev objects) [] in
    let out s = output_string main_k2 s
    in out "(:declaration \"#include \\\"";
       out path_include; out "z2k2.h\\\"\")\n";
       if not boehm then out "(:declare :static :defun gcvars ())\n";
       out "(:declare :static :defun objects ())\n";
       out "(:declaration \"#include \\\"";
       out path_include; out "main.h\\\"\")\n";
       if not boehm then
       begin out "(:defun gcvars ()\n";
             do_list (fun modu -> out "(vars_of_"; out modu; out ")\n") 
                     modules;
             do_list (fun obj -> out "(vars_of_"; 
                                 out (basename obj); 
                                 out ")\n") 
                     objects;
             out ")\n"
       end;
       out "(:declaration \"struct {int tag; char *s;} $[] = {";
       do_hashtable (fun (t,s) -> out (string_of_int t);
				  out ",\\\"";
				  out s;
				  out "\\\",") exc_table;
       out "0,\\\"\\\"};\" *exctable*)\n";
       out "(:defun objects ()\n";
       do_list (fun modu -> out "(init_of_"; out modu; out ")\n") modules;
       do_list (fun obj -> out "(init_of_"; out (basename obj); out ")\n") 
               objects;
       do_list (fun modu -> out "(main_of_"; out modu; out ")\n") modules;
       do_list (fun obj -> out "(main_of_"; out (basename obj); out ")\n") 
               objects;
(*       out "(io:exit*1 (:expression \"(obj_t)0\"))\n"; *)
         out ")\n";
       close_out main_k2;
       print_string z2k2_main; 
       map (fun modu -> print_string (" " ^ (find_in_path (modu ^ ".o")))) 
	   (rev modules);
       print_newline ()
;;
